/*
 * pftp -- sends files from host to host through free choosable ports
 *
 * Copyright (C) 1996-1999 Ben Schluricke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the emplied warranty of MERCHANT-
 * ABILITY OF FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *    Written by Ben Schluricke
 *    E-Mail:    bhor0533@lehr.chem.TU-Berlin.DE
 *
 * This program is dedicated to my girl-friend, Heather O'Rourke.
 *
 *
 */
#ifdef USE_POSIX_THREAD
#define _REENTRANT
#include <pthread.h>
#endif
#include <sys/types.h>
#include <pwd.h>
#include "main.h"

extern void set_tty(int);
extern void shost(int, int *, short);
extern void lhost(char *, int *, short);
extern short get_var_from_pftprc(const char *, char *, short);
extern void pmanager(short);
extern void warranty(void);
extern void display_pa(char **, char *, char *, int);
extern short mycmp(char *, char *);
int printlist(int, char *, short);
void create_passwd(void);
int getoffset(char *);
long offset[SONAME];  /* hold all offsets of the beginning of files */
static int extract_val_d(double *, char *, int);
static int extract_val_i(unsigned int *, char *, int);
static int extract_val_li(unsigned long *, char *, int);

#define HELPLINES 95

void help() {
   char *firstline=NULL;
   char *helpstr[HELPLINES+1] = {
      " Server: pftp [port] [OPTIONS]                                   \n",
      "                                                                 \n",
      "    -         Remote stdin on client side is written to stdout.  \n",
      "              Files are written to disk.                         \n",
      "    -a        Overwrite all existing files.                      \n",
      "    -ANUM     Accept NUM clients at once (default: 255).         \n",
      "    -b[NUM]   Same as `-' except that stdin and stdout is        \n",
      "              buffered. NUM is the size in bytes of the buffers. \n",
      "              The buffer size defaults to 4096 bytes.            \n",
      "    -bb       Same as `-b32768'.                                 \n",
      "    -B[NUM]   Set the net buffer size in bytes to the maximum    \n",
      "              value specified by NUM, or if omitted the value of \n",
      "              the environment variable PFTPNETBUF is taken.      \n",
      "    -BB       Same as `-B65536'.                                 \n",
      "    -c hosts  Accept specified clients only.  Hosts beginning    \n",
      "              with a `.' are supposed to be domain names.        \n",
      "              If hosts are omitted the variable PFTPCLIENTS      \n",
      "              which may hold host and domain names is read.      \n",
      "    -d        Start pftp as a daemon. The number of the filter   \n",
      "              to be used is specified by the client.  The filter \n",
      "              is read from user's pftp resource file.            \n",
      "    -f[NUM]   Use filter NUM specified in the variable           \n",
      "              PFTPSFILTER.  NUM defaults to 1.                   \n",
      "    -i        Start pftp from `/etc/inetd.conf'.  The number of  \n",
      "              the filter to be used is specified by the client.  \n",
      "              The filter is read from user's pftp resource file. \n",
      "    -m[NUM]   Receive multicasted UDP datagrams.  The payload    \n",
      "              length is set to NUM (defaults to 1472).           \n",
#if defined IPV6_ADD_MEMBERSHIP || (defined IP_ADD_MEMBERSHIP && !defined HAVE_INET6)
      "              This binary supports multicasting, your system too?\n",
#else
      "              This binary does *not* support multicasting.       \n",
#endif
      "    -mg GROUP Join multicast group GROUP.                        \n",
      "    -mi IADDR Use interface IADDR for multicasted datagrams.     \n",
      "    -o        Accept only one connection at a time.              \n",
      "    -p[NUM]   Send all received files to standard output.  NUM is\n",
      "              the size of the output buffer.  If NUM is omitted  \n",
      "              it is linebuffered by default.                     \n",
      "    -pp       Same as `-p32768'.                                 \n",
      "    -q        Run silently--existing files are skipped by default\n",
      "    -r        Receive directories recursively.                   \n",
      "    -s        TCP: Skip all existing files.  UDP: Display bytes. \n",
      "    -T        Net performance test (requires `-b').              \n",
      "    -u[NUM]   Use UDP.  NUM is the string size (default: 1472).  \n",
      "    -v        Be a little more verbose.                          \n",
      "    -x        Exit after first connection.  Implies `-o'.        \n",
      "                                                                 \n",
      "                                                                 \n",
      " Client: pftp [port] [OPTIONS] [-l|-NUM| hostname] [files]       \n",
      "                                                                 \n",
      "    -         Send stdin unbuffered.  Standard input is          \n",
      "              incompatible with option `-l'.                     \n",
      "    -aN1aN2   Begin sending at the N1th byte of the first file...\n",
      "              See the manual page for further notices.           \n",
      "    -b[NUM]   Same as `-' except that stdin is buffered.         \n",
      "              NUM is the size in bytes of the buffer.            \n",
      "              The buffer size defaults to 4096 bytes.            \n",
      "    -bb       Same as `-b32768'.                                 \n",
      "    -B[NUM]   Set the net buffer size in bytes to the maximum    \n",
      "              value specified by NUM, or if omitted the value of \n",
      "              the environment variable PFTPNETBUF is taken.      \n",
      "    -BB       Same as `-B65536'.                                 \n",
      "    -d[=DIR]  Send data to the daemon.  The files are uploaded   \n",
      "              in the directory DIR on the remote host.           \n",
      "    -e        Create a normal Unix password with crypt()         \n",
      "    -f[NUM]   Use filter NUM specified in variable               \n",
      "              PFTPCFILTER.  NUM defaults to 1.                   \n",
      "    -i[=DIR]  Send data to the inetd server.  The files are      \n",
      "              uploaded in the directory DIR on the remote host.  \n",
      "    -j        Read file and directory names from standard input  \n",
      "    -l        Select hostname from list and send the files       \n",
      "              specified on command line.                         \n",
      "    -L        Do not follow symbolic links. Instead send a script\n",
      "              file PFTP_MAKE_LINKS which may create all links.   \n",
      "    -m[NUM]   Receive multicasted UDP datagrams.  The payload    \n",
      "              length is set to NUM (defaults to 1472).           \n",
      "    -mi IADDR Use interface IADDR for multicasted datagrams.     \n",
      "    -ml       Multicasted datagrams will be looped back.         \n",
      "    -mtNUM    Time to live (IPv4) resp. hop limit (IPv6).        \n",
      "    -M[=FILE] In combination with `-n': Add message file FILE or \n",
      "              if FILE is not given write the file interactively. \n",
      "    -N        Peek for new data in the upload directory.         \n",
      "    -n        Start the pftp file and directory manager.         \n",
      "    -nn       Send data to a friend and ask for subject.         \n",
      "    -n[=SUBJ] Send data [of subject SUBJ] to a friend.           \n",
      "    -NUM      NUM is the number of the host in your pftp resource\n",
      "              file.  Try `pftp -l .' to see the numbers.         \n",
      "    -P[NUM]   Set datagram priority to NUM (for IPv6).           \n",
#if defined HAVE_INET6
      "              This binary supports IPv4 and IPv6.                \n",
#else
      "              This binary supports IPv4 only.                    \n",
#endif
      "    -q        Run silently.                                      \n",
      "    -Q        Used by the pftp front end, qpftp.                 \n",
      "    -r        Send directories recursively.                      \n",
      "    -SNUM     UDP: Drop every NUMth datagram.                    \n",
      "    -T[NUM]   Net performance test with NUM bytes to send.       \n",
      "    -u[NUM]   Use UDP.  NUM is the string size (default: 1472).  \n",
      "    -v        Be a little more verbose.                          \n",
      "    -w[NUM]   Set the bandwidth in bytes to NUM, or if omitted to\n",
      "              the value of the environment variable PFTPBANDWID. \n",
      "    -W        Display warranty.                                  \n"
   };
   
   if (isatty(1)) {
      char *ctmp=NULL;
      MEM_CHECK((firstline = (char *)calloc(SONAME, sizeof(char))));
      sprintf(firstline, "*** Port-FTP Version %s Copyright (C) 1996-1999 Ben Schluricke ***", VERSION); 
      for (ctmp=firstline; *ctmp; ctmp++);
      for (; ctmp - firstline < _WINCOLS_; ctmp++) *ctmp = '*';
      *(ctmp-1) = ' ';
      *ctmp = '\0';
      fprintf(stdout, "[?25l[H[J");
      set_tty(2);
      display_pa(helpstr, firstline, "                                                                            ", HELPLINES);  
      set_tty(0);
      fprintf(stderr, "[?25h\r                                                                       \r");
      free(firstline);
   }
   else {
      int j=0;
      fprintf(stdout, "\n*** Port-FTP Version %s ", VERSION); 
      fprintf(stdout, "Copyright (C) 1996-1999 Ben Schluricke ***\n\n");
      for (j=0; j <= HELPLINES; j++) fputs(*(helpstr+j), stdout);
   }
   exit(0);
}

int argum(int argc, char **argv, int *portn)
{
   int _ISDIGIT_=0;
   FILE *mfs;
   (*statstr)->usefilter = 0;
   *offset = -1;

   /*
    * Set the default maximum package size.
    */
   (*statstr)->bsize = BUFSIZE;

   /*
    * Please don't change this option.
    */
   if (argc == 2 && (strcmp(*(argv+1), "-H") == 0)) {
      char *blankline=NULL, *ctmp=NULL;
      MEM_CHECK((blankline = (char *)calloc(SONAME, sizeof(char))));
      for (ctmp=blankline; ctmp - blankline < (_WINCOLS_-45)/2; ctmp++) *ctmp = ' ';
      *ctmp='\0';
      fprintf(stderr, "[?25l[31m[H[J[%dH%s", _WINROWS_/2, blankline);
      free(blankline);
      fprintf(stderr, "This program is dedicated to Heather O'Rourke");
      fprintf(stderr, "[0m[%dH", _WINROWS_);
      set_tty(2); fgetc(stdin); set_tty(0);
      fprintf(stderr, "[?25h[H[J");
      exit(0);
   }
   else if (argc == 2 && !mycmp(*(argv+1), "-n")) {
      pmanager(0);
      exit(0);
   }
   else if (argc == 2 && !mycmp(*(argv+1), "-N")) {
      pmanager(1);
      exit(0);
   }
   else if (argc == 2 && (mycmp(*(argv+1), "-W") == 0)) {
      fprintf(stderr, "[?25l");
      set_tty(2);
      warranty();
      set_tty(0);
      fprintf(stderr, "[?25h\r");
      exit(0);
   }
   else if (argc == 1) {
      (*statstr)->_OPTIONS_=1;
      *portn=-1;
      lhost(NULL, portn, 0);
      return 1;
   }
   else if (argc == 2 && (atoi(*(argv+1)) && strlen(*(argv+1)) > 6) && **(argv+1) != '-') {
      return 1;
   }
   else if (*argv[1] == '-' || argc > 1) {
      int i, j, c, _NEXT_, cl=0;
      char *gtmp=NULL;
      short clhosts=0;

      if (!strcmp(*(argv+1), "-h")) help();
      else if (**(argv+1) == '-' || !atoi(*(argv+1)) || strlen(*(argv+1)) > 5) {
         *portn=-1;
         lhost(NULL, portn, 0);
         (*statstr)->_OPTIONS_ = 1;
      }
      else {
         if (!(*portn = atoi(*(argv+1)))) *portn = 0;
         (*statstr)->_OPTIONS_ = 2;
      }

      while ((*statstr)->_OPTIONS_ < argc && *argv[(*statstr)->_OPTIONS_] == '-') {
         _NEXT_ = 1;
         clhosts = 0;
         if (!mycmp(argv[(*statstr)->_OPTIONS_], "-")) {
            (*statstr)->_STANDARD_INPUT_ = BIT_ONE;
            (*statstr)->_STDIN_BUFSIZ_ = 0;
         }
         for (i=1; _NEXT_ && (c = *(argv[(*statstr)->_OPTIONS_]+i)); i++) {
            switch(c) {
               case 'c':
                  /* get host names of accepted clients */
                  if (argc > (*statstr)->_OPTIONS_ + 1) {
                     if (!getenv("PFTPCLIENTS") && *argv[(*statstr)->_OPTIONS_+1] == '-') {
                        MEM_CHECK((gtmp = (char *)calloc(SONAME, sizeof(char))));
                        if (!get_var_from_pftprc("PFTPCLIENTS", gtmp, 0)) help();
                     }
                     else if (*argv[(*statstr)->_OPTIONS_+1] != '-') clhosts = 1;
                  }
                  else if (argc == (*statstr)->_OPTIONS_ + 1) {
                     if (!getenv("PFTPCLIENTS")) {
                        MEM_CHECK((gtmp = (char *)calloc(SONAME, sizeof(char))));
                        if (!get_var_from_pftprc("PFTPCLIENTS", gtmp, 0)) help();
                     }
                  }
                  if (!*_CLIENTHOSTNAME_ && isatty(2)) {
                     fprintf(stderr, "\n*** Connections are accepted from the following host(s) ***\n\n");
                  }
                  for (cl=0; cl < MAXCLIENTHOSTS && *(_CLIENTHOSTNAME_+cl); cl++);

                  /*
                   * Read from command line.
                   */
                  if (clhosts) {
                     for (; ++(*statstr)->_OPTIONS_ < argc && *argv[(*statstr)->_OPTIONS_] != '-'; cl++) {
                        MEM_CHECK((*(_CLIENTHOSTNAME_+cl) = (char *)calloc(SONAME, sizeof(char))));
                        printlist(cl, argv[(*statstr)->_OPTIONS_], 0);
                     }
                     if (!*_CLIENTHOSTNAME_) help();
                     _NEXT_ = 0;
                     (*statstr)->_OPTIONS_--;
                     continue;
                  }
                  /*
                   * Read from environment variable or resource file.
                   */
                  else if (getenv("PFTPCLIENTS") || gtmp) {
                     char *tmp=NULL;
                     if (getenv("PFTPCLIENTS")) {
                        if (!gtmp) MEM_CHECK((gtmp = (char *)calloc(SONAME, sizeof(char))));
                        strcpy(gtmp, getenv("PFTPCLIENTS"));
                     }
                     tmp = gtmp;
                     while (*tmp == ' ' || *tmp == '*') tmp++;
                     for (; *tmp; cl++) {
                        MEM_CHECK((*(_CLIENTHOSTNAME_+cl) = (char *)calloc(SONAME, sizeof(char))));
                        tmp += printlist(cl, tmp, 0);
                        while (*(++tmp) == ' ' || *tmp == '*');
                     }
                     if (!*_CLIENTHOSTNAME_) help();
                     free(gtmp);
                  }
                  else help();
                  break;
               case 'w':
                  /*
                   * Get the size of the bandwidth
                   */
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                     i = extract_val_li((unsigned long *)&((*statstr)->_BANDWIDTH_), argv[(*statstr)->_OPTIONS_], i);
                  }
                  else if (!getenv("PFTPBANDWID")) {
                     MEM_CHECK((gtmp = (char *)calloc(SONAME, sizeof(char))));
                     if (!get_var_from_pftprc("PFTPBANDWID", gtmp, 0)) {
                        fprintf(stderr, "** Please set the variable PFTPBANDWID!\n");
                        exit(1);
                     }
                     (*statstr)->_BANDWIDTH_ = (size_t) atol(gtmp);
                     free(gtmp);
                  }
                  else (*statstr)->_BANDWIDTH_ = (size_t) atol(getenv("PFTPBANDWID"));
                  if (((*statstr)->_BANDWIDTH_ < BUFSIZE) && !(*statstr)->_STDIN_BUFSIZ_ && !(*statstr)->_STANDARD_INPUT_) {
                     (*statstr)->_STDIN_BUFSIZ_ = (*statstr)->_BANDWIDTH_;
                  }
                  break;
               case 'B':
                  /*
                   * Get the size of the net buffer.
                   */
                  cl = 0;
                  if (!(*statstr)->use_udp) {
                     if (argv[(*statstr)->_OPTIONS_][i+1] == 'B') {
                        cl = 1;
                        i++;
                     }
                     if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                        i = extract_val_i((unsigned int *)&((*statstr)->bsize), argv[(*statstr)->_OPTIONS_], i);
                     }
                     else if (cl) (*statstr)->bsize = BIG_BUFSIZE;
                     else if (!getenv("PFTPNETBUF")) {
                        MEM_CHECK((gtmp = (char *)calloc(SONAME, sizeof(char))));
                        if (!get_var_from_pftprc("PFTPNETBUF", gtmp, 0)) {
                           fprintf(stderr, "** Please set the variable PFTPNETBUF!\n");
                           exit(1);
                        }
                        (*statstr)->bsize = (size_t) atol(gtmp);
                        free(gtmp);
                     }
                     else (*statstr)->bsize = (size_t) atol(getenv("PFTPNETBUF"));
                  }
                  else {
                     fprintf(stderr, "** There is no net buffer size in UDP.\n");
                     exit(1);
                  }
                  break;
               case 'p':
                  /*
                   * Force sending all received data to stdout.
                   */
                  (*statstr)->FORCE_STDOUT = BIT_ONE;
               case 'b':
                  /*
                   * Get and set string size.
                   */
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) { 
                     i = extract_val_i((unsigned int *)&((*statstr)->_STDIN_BUFSIZ_), argv[(*statstr)->_OPTIONS_], i);
                  } 
                  else if (argv[(*statstr)->_OPTIONS_][i+1] == 'p' || argv[(*statstr)->_OPTIONS_][i+1] == 'b') {
                     (*statstr)->_STDIN_BUFSIZ_ = DEFAULT_STDOUT_BUFSIZ;
                     i++;
                  }
                  else {
                     (*statstr)->_STDIN_BUFSIZ_ = c == 'p' ? 0: DEFAULT_STDIN_BUFSIZ;
                  }
               case '-':
                  if (c == 'p') break;
                  (*statstr)->_STANDARD_INPUT_ = BIT_ONE;
                  if (c == '-') (*statstr)->_STDIN_BUFSIZ_ = 0;
                  break;
               case 'j':
                  (*statstr)->READFROMSTDIN = 1;
                  break;
               case 'e':
                  create_passwd();
                  break;
               case 'f':
                  if ((cl = argv[(*statstr)->_OPTIONS_][i+1]) == 'S' || cl == 'C') {
                     char *tmp=NULL;
                     /*
                      * List all Server/Client filters.
                      */
                     if (cl == 'S') {
                        tmp = getenv("PFTPSFILTER");
                        if (!tmp) {
                           MEM_CHECK((tmp = (char *)calloc(SONAME, sizeof(char))));
                           if (!get_var_from_pftprc("PFTPSFILTER", tmp, 0)) {
                              free(tmp);
                              tmp = NULL;
                           }
                        }
                     }
                     else {
                        tmp = getenv("PFTPCFILTER");
                        if (!tmp) {
                           MEM_CHECK((tmp = (char *)calloc(SONAME, sizeof(char))));
                           if (!get_var_from_pftprc("PFTPCFILTER", tmp, 0)) {
                              free(tmp);
                              tmp = NULL;
                           }
                        }
                     }
                     if (!tmp) {
                        fprintf(stderr, "** %s is not set.\n", cl == 'S'?"PFTPSFILTER":"PFTPCFILTER");
                        exit(1);
                     }
                     for (j=0;*(tmp+j) == ' ' && *(tmp+j); j++);
                     if (!*(tmp+j)) {
                        fprintf(stderr, "** %s is empty.\n", cl == 'S'?"PFTPSFILTER":"PFTPCFILTER");
                     }
                     else {
                        int skip_space=1;
                        fprintf(stderr, "\n* List of %s filters:\n\n", cl == 'S'?"server":"client");
                        fprintf(stderr, "      1) ");
                        for (i=0,j=2; *(tmp+i); i++) {
                           if (*(tmp+i) == ':') {
                              fprintf(stderr, "\n      %d) ", j++);
                              skip_space = 1;
                           }
                           else {
                              if (*(tmp+i) == ' ' || *(tmp+i) == '\t') {
                                 while(*(tmp+i) == ' ' || *(tmp+i) == '\t') i++;
                                 if (!skip_space) i--;
                              }
                              skip_space = 0;
                              fputc(*(tmp+i), stderr);
                           }
                        }
                        fprintf(stderr, "\n\n");
                     }
                     exit(0);
                  }
                  else {
                     if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                        i = extract_val_i((unsigned int *)&((*statstr)->usefilter), argv[(*statstr)->_OPTIONS_], i);
                     }
                     if (!(*statstr)->usefilter) (*statstr)->usefilter=1;
                  }
                  break;
               case 'M':
                  (*statstr)->delete_info = 0;
                  MEM_CHECK((data_info = (char *)calloc(SONAME, sizeof(char))));
                  if (*(argv[(*statstr)->_OPTIONS_]+i+1) == '=') {
                     strcpy(data_info, (argv[(*statstr)->_OPTIONS_]+i+2));
                     for (; *(argv[(*statstr)->_OPTIONS_]+i+1); i++) {
                        *(argv[(*statstr)->_OPTIONS_]+i+1) = '\0';
                     }
                     _NEXT_ = 0;
                  }
                  else {
                     (*statstr)->delete_info = 1;
                     *data_info = '\0';
                  }
                  break;
               case 'n':
                  if (*(argv[(*statstr)->_OPTIONS_]+i+1) == '=') {
                     MEM_CHECK((data_subject = (char *)calloc(SONAME, sizeof(char))));
                     strcpy(data_subject, (argv[(*statstr)->_OPTIONS_]+i+2));
                     strcat(data_subject, "\n");
                     /*
                      * Nobody else should see the subject.
                      */
                     for (; *(argv[(*statstr)->_OPTIONS_]+i+1); i++) {
                        *(argv[(*statstr)->_OPTIONS_]+i+1) = '\0';
                     }
                     _NEXT_ = 0;
                  }
                  else if (*(argv[(*statstr)->_OPTIONS_]+i+1) == 'n') {
                     MEM_CHECK((data_subject = (char *)calloc(SONAME, sizeof(char))));
                     fprintf(stdout, "Subject: ");
                     fgets(data_subject, SONAME, stdin);
                     i++;
                  }
                  MEM_CHECK(((*statstr)->from = (char *)calloc(SONAME, sizeof(char))));
                  if (getpwuid(getuid())) {
                     strcpy((*statstr)->from, (char *)getpwuid(getuid())->pw_name);
                  }
                  else {
                     if (stderr) fprintf(stderr, "** Who are you?!\n");
                     exit(1);
                  }
                  if (!(*statstr)->_PFTP_DAEMON_) {
                     (*statstr)->_PFTP_DAEMON_ = BIT_TWO;
#ifdef INETD_PORT
                     *portn = INETD_PORT;
#else
                     *portn = 26;
#endif
                  }
                  break;
               case 'i':
                  /*
                   * Exit after first connection.
                   * This is meant for `inetd.conf'.
                   */
                  (*statstr)->_PFTP_DAEMON_ = BIT_TWO;
#ifdef INETD_PORT
                  *portn = INETD_PORT;
#else
                  *portn = 26;
#endif
               case 'd':
                  /*
                   * Run as long as to be killed.
                   */
                  if (c == 'd') (*statstr)->_PFTP_DAEMON_ = BIT_ONE;
                  client_destdir = NULL;
                  MEM_CHECK((client_destdir = (char *)calloc(LONAME, sizeof(char))));
                  if (*(argv[(*statstr)->_OPTIONS_]+i+1) == '=') {
                     strcpy(client_destdir, (argv[(*statstr)->_OPTIONS_]+i+2));
                     /*
                      * Nobody else should see the destination directory.
                      */
                     for (; *(argv[(*statstr)->_OPTIONS_]+i+1); i++) {
                        *(argv[(*statstr)->_OPTIONS_]+i+1) = '\0';
                     }
                     _NEXT_ = 0;
                  }
                  else strcpy(client_destdir, "~");
                  break;
               case 'A':
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) { 
                     i = extract_val_i((unsigned int *)&(*statstr)->maxclients, argv[(*statstr)->_OPTIONS_], i);
                  }
                  else {
                     fprintf(stderr, "** The number of accepted clients can be 1 to %d.\n", MAXCLIENTS);
                     exit(1);
                  }
                  break;
               case 'a':
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) { 
                     if (*offset < 0) i =+ getoffset((argv[(*statstr)->_OPTIONS_]+i+1))+2;
                     else help();
                  }
                  else {
                     (*statstr)->OVERWRITE = BIT_TWO;
                     (*statstr)->_SKIP_ = 0;
                  }
                  break;
               case 'r':
                  (*statstr)->_RECURS_ = 1;
                  break;
               case 'o':
                  (*statstr)->single_connection = 1;
                  break;
               case 'l':
                  if (!isatty(2)) {
                     fprintf(stderr, "** Sorry, '-l' needs a tty for output.\n");
                     exit(1);
                  }
                  else if (_ISDIGIT_ == 1) help();
                  else _ISDIGIT_ = 2;
                  break;
               case 'L':
                  MEM_CHECK(((*statstr)->SENDMAKELINKS = (char *)calloc(SONAME, sizeof(char))));
                  MEM_CHECK((gtmp = (char *)calloc(SONAME, sizeof(char))));
                  if (getenv("TMP")) {
                     int len=0;
                     strcpy(gtmp, getenv("TMP"));
                     len = strlen(gtmp);
                     if (*(gtmp+len-1) != '/') {
                        *(gtmp+len) = '/';
                        *(gtmp+len) = '\0';
                     }
                     strcat(gtmp, PFTPMAKELINKS);
                  }
                  sprintf((*statstr)->SENDMAKELINKS, "%s.%d", gtmp, getpid());
                  if ((mfs = fopen((*statstr)->SENDMAKELINKS, "w+")) == NULL) {
                     fprintf(stderr, "Can't open file %s to temporary store the links.\n", (*statstr)->SENDMAKELINKS);
                     exit(1);
                  }
                  fputs(PFTPLINKSCRIPTINIT, mfs);
                  fclose(mfs);
                  free(gtmp);
                  break;
               case 's':
                  (*statstr)->show_bytes = BIT_ONE;
                  (*statstr)->_SKIP_ = BIT_ONE;
                  (*statstr)->OVERWRITE = 0;
                  break;
               case 'q':
                  slfp = NULL;
                  break;
               case 'Q':
                  (*statstr)->QT_FRONTEND = 1;
                  (*statstr)->READFROMSTDIN = 1;
                  break;
#ifdef HAVE_INET6
               case 'P':
                  /*
                   * Set datagram priority for IPv6.
                   */
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                     i = extract_val_i((unsigned int *)&((*statstr)->priority), argv[(*statstr)->_OPTIONS_], i);
                  }
                  else (*statstr)->priority = PFTP_DEFAULT_PRIORITY;
                  break;
#endif
               case 'S':
                  /*
                   * Sets the period of dropping datagrams.
                   */
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                     i = extract_val_li((unsigned long *)&((*statstr)->throw_away), argv[(*statstr)->_OPTIONS_], i);
                  }
                  else help();
                  break;
               case 'T':
                  /*
                   * Get the number of bytes to send during the test.
                   */
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                     i = extract_val_d((double *)&((*statstr)->testbsize), argv[(*statstr)->_OPTIONS_], i);
                  }
                  else (*statstr)->testbsize = 1L;
                  break;
               case 'x':
                  (*statstr)->exit_afc = 1;
                  break;
#if defined IPV6_ADD_MEMBERSHIP || (defined IP_ADD_MEMBERSHIP && !defined HAVE_INET6)
               case 'm':
                  /*
                   * Receive/Send multicasted datagrams.
                   */
                  (*statstr)->multicast = 1;
                  if (argv[(*statstr)->_OPTIONS_][i+1] && !isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                     switch (argv[(*statstr)->_OPTIONS_][i+1]) {
                        case 'g':
                           if (argv[(*statstr)->_OPTIONS_+1] && *argv[(*statstr)->_OPTIONS_+1] != '-') {
                              MEM_CHECK(((*statstr)->_MULTICASTGROUP_ = (char *)calloc(SONAME, sizeof(char))));
                              (*statstr)->_OPTIONS_++;
                              strcpy((*statstr)->_MULTICASTGROUP_, argv[(*statstr)->_OPTIONS_]);
                              if (!(*statstr)->use_udp) (*statstr)->bsize = DEFAULT_UDP_SIZE;
                              (*statstr)->use_udp = BIT_ONE;
                              _NEXT_ = 0;
                              continue;
                           }
                           else help();
                           break;
                        case 'i':
                           if (argv[(*statstr)->_OPTIONS_+1] && *argv[(*statstr)->_OPTIONS_+1] != '-') {
                              MEM_CHECK(((*statstr)->_INTERFACE_ = (char *)calloc(SONAME, sizeof(char))));
                              (*statstr)->_OPTIONS_++;
                              strcpy((*statstr)->_INTERFACE_, argv[(*statstr)->_OPTIONS_]);
                              if (!(*statstr)->use_udp) (*statstr)->bsize = DEFAULT_UDP_SIZE;
                              (*statstr)->use_udp = BIT_ONE;
                              _NEXT_ = 0;
                              continue;
                           }
                           else help();
                           break;
                        case 'l':
                           (*statstr)->mloop = 1;
                           i++;
                           break;
                        case 't':
                           i += 2;
                           if (argv[(*statstr)->_OPTIONS_][i]) {
                              (*statstr)->mttl = atoi(argv[(*statstr)->_OPTIONS_]+i);
                              i += strlen(argv[(*statstr)->_OPTIONS_]+i) - 1;
                           }
                           else help();
                           break;
                     }
                     if (!(*statstr)->use_udp) (*statstr)->bsize = DEFAULT_UDP_SIZE;
                     (*statstr)->use_udp = BIT_ONE;
                     break;
                  }
#endif
               case 'u':
                  if (isdigit(argv[(*statstr)->_OPTIONS_][i+1])) {
                     i = extract_val_i((unsigned int *)&((*statstr)->bsize), argv[(*statstr)->_OPTIONS_], i);
                  }
                  else (*statstr)->bsize = DEFAULT_UDP_SIZE;
                  (*statstr)->use_udp = BIT_ONE;
                  break;
               case 'v':
                  (*statstr)->verbose = BIT_ONE;
                  break;
               case 'h':
                  help();
                  break;
               default:
                  if (_ISDIGIT_ > 1 || !isdigit(c)) {
                     fprintf(stderr, "**\n** unknown option `%c'\n**\n", c);
                     help();
                  }
                  else if (!_ISDIGIT_) {
                     lhost((argv[(*statstr)->_OPTIONS_]+i), portn, 0); 
                     _ISDIGIT_ = 1;
                  }
                  break;
            }
         }
         if (i == 1) (*statstr)->_STANDARD_INPUT_ = 1;
         (*statstr)->_OPTIONS_++;
      }
   }
   else if (argc < 3) {
      help();
   }

   if (!(*statstr)->_STDIN_BUFSIZ_ && !(*statstr)->_STANDARD_INPUT_ && !(*statstr)->FORCE_STDOUT) (*statstr)->_STDIN_BUFSIZ_ = DEFAULT_STDOUT_BUFSIZ;

   /*
    * Indentify and check.
    */
   if(_ISDIGIT_) (*statstr)->_OPTIONS_--;
   if ((*statstr)->READFROMSTDIN) {
      (*statstr)->_STANDARD_INPUT_ = 0;
   }
   if ((*statstr)->testbsize && !(*statstr)->_STANDARD_INPUT_ \
      && !(*statstr)->_BANDWIDTH_ && !(*statstr)->use_udp) {
      fprintf(stderr, "\n******************** Net performance test ************************\n");
      fprintf(stderr, "**                                                              **\n");
      fprintf(stderr, "** You must specify `-b[NUM]' for a TCP based test respectively **\n");
      fprintf(stderr, "** `-u[NUM] for a UDP based test with NUM as string size to be  **\n");
      fprintf(stderr, "** sent/received.  You can additionally use option `-B[NUM]'    **\n");
#ifdef HAVE_INET6
      fprintf(stderr, "** to specify the net buffer size, set the bandwidth with       **\n");
      fprintf(stderr, "** `-w[NUM]', and set the datagram priority with `-P[NUM]'.     **\n");
#else
      fprintf(stderr, "** to specify the net buffer size and set the bandwidth with    **\n");
      fprintf(stderr, "** `-w[NUM]'.                                                   **\n");
#endif
      fprintf(stderr, "**                                                              **\n");
      fprintf(stderr, "***************************************************************\n\n");
      exit(1);
   }
   if ((*statstr)->testbsize && (*statstr)->use_udp) (*statstr)->_STANDARD_INPUT_ = BIT_ONE;
   if(((*statstr)->OVERWRITE && (*statstr)->_SKIP_) || ((*statstr)->OVERWRITE && _ISDIGIT_)) help();
   /*
    * If the program is running as the server return 1 else 
    * if the program is running as the client return 0.
    * Print help a message if an error occurs.
    */
   else if ((((argc == 2) && (!isalpha(argv[1][0]))) \
        || ((*statstr)->_OPTIONS_ == argc)) && (!_ISDIGIT_) \
        && (!(*statstr)->_BANDWIDTH_) && (*offset < 0) \
        && !(*statstr)->READFROMSTDIN \
        && !(*statstr)->SENDMAKELINKS)
        /*
         * Program is running as the server.
         */
        return 1;
   else if (((*statstr)->OVERWRITE || (*statstr)->_SKIP_ || *_CLIENTHOSTNAME_) \
        || (_ISDIGIT_ > 1 && (*statstr)->_STANDARD_INPUT_) \
        || ((*statstr)->_RECURS_ && (*statstr)->_STANDARD_INPUT_) \
        || ((*statstr)->_RECURS_ && !_ISDIGIT_ && !(*statstr)->READFROMSTDIN && (*statstr)->_OPTIONS_ == argc - 1) \
        || ((*statstr)->_STANDARD_INPUT_ && argv[argc-2][0] != '-') \
        || (!(*statstr)->_STANDARD_INPUT_ && !(*statstr)->READFROMSTDIN && argv[argc-1][0] == '-' \
        && !((*statstr)->use_udp && (*statstr)->testbsize)) \
        || ((*statstr)->_STANDARD_INPUT_ && argv[argc-1][0] != '-'  && _ISDIGIT_) \
        || ((*statstr)->exit_afc) || ((*statstr)->FORCE_STDOUT) \
        || (argc < 3))
           help();
   else if (_ISDIGIT_ > 1) lhost("l", portn, 0);

   if ((!(*statstr)->_STANDARD_INPUT_ ||_ISDIGIT_ > 1) && (*statstr)->_OPTIONS_ == argc - 1) {
      fprintf(stderr, "** No files and no standard input option specified!\n");
      exit(1);
   }

   /*
    * Program is running as the client.
    */
   return 0;
}


int printlist(int cl, char *chost, short strnum)
{
   char *clienthost = NULL;
   struct hostent *hp;
   short _ISNUM_=1;
   char *tmphost=NULL;
   int inum, j, tmp;

   MEM_CHECK((clienthost = (char *)calloc(SONAME, sizeof(char))));
   strcpy(clienthost, chost);
   for (j=0; *(clienthost+j) != ' ' && *(clienthost+j); j++)
      if (!isdigit(*(clienthost+j))) _ISNUM_=0;
   *(clienthost+j) = '\0';
   tmp = j;
   if (!_ISNUM_) {
      tmphost = clienthost;
   }
   else {
      int dummy;
      shost(atoi(clienthost), &dummy, strnum);
      tmphost = (*statstr)->_HOSTNAME_;
   }
   inum = inet_addr(tmphost);
   if (*tmphost != '.') {
      if ((hp = (struct hostent *)gethostbyname(tmphost)) != NULL);
      else if ((hp = gethostbyaddr((char *)&inum, sizeof(inum), AF_INET)) == 0) {
         fprintf(stderr, "** %s: unknown host.\n", tmphost);
         exit(1);
      }
      if (isatty(2)) fprintf(stderr, "   %d) %s\n", cl+1, hp->h_name);
      for (j=0; *(hp->h_name+j); j++) *(*(_CLIENTHOSTNAME_+cl)+j) = tolower(*(hp->h_name+j));
   }
   else {
      if (isatty(2)) fprintf(stderr, "   %d) *%s\n", cl+1, tmphost);
      for (j=0; *(tmphost+j); j++) *(*(_CLIENTHOSTNAME_+cl)+j) = tolower(*(tmphost+j));
   }
   free(clienthost);
   return tmp;
}

int getoffset(char *z)
{
   char *tmpz = z;
   long *t = offset;

   for (*t=0L; *z; z++) {
      if (isdigit(*z)) *t = (*t * 10) + (long) (*z - '0');
      else if (*z == 'a' && isdigit(*(z+1))) {
         t++;
         *t = 0L;
      }
      else break;
   }
   t++;
   *t = -1;

   return (int) (z - tmpz -1);
}

static int extract_val_d(double *num_to_set, char *str, int i)
{
      int j=0;

      for (j=i+1,*num_to_set=0; isdigit(*(str+j)); j++) {
         if (*num_to_set) *num_to_set *= (double)10;
         *num_to_set += (double)*(str+j) - '0';
      }
      return j - 1;
}

static int extract_val_i(unsigned int *num_to_set, char *str, int i)
{
      int j=0;

      for (j=i+1,*num_to_set=0; isdigit(*(str+j)); j++) {
         if (*num_to_set) *num_to_set *= (unsigned int)10;
         *num_to_set += (unsigned int)*(str+j) - '0';
      }
      return j - 1;
}

static int extract_val_li(unsigned long *num_to_set, char *str, int i)
{
      int j=0;

      for (j=i+1,*num_to_set=0; isdigit(*(str+j)); j++) {
         if (*num_to_set) *num_to_set *= (unsigned long)10;
         *num_to_set += (unsigned long)*(str+j) - '0';
      }
      return j - 1;
}
